package fr.eni.ep3jasp.doc;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.hateoas.MediaTypes;
import org.springframework.restdocs.JUnitRestDocumentation;
import org.springframework.restdocs.payload.JsonFieldType;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;
import org.springframework.web.context.WebApplicationContext;

import javax.servlet.RequestDispatcher;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import static org.hamcrest.Matchers.is;
import static org.hamcrest.Matchers.notNullValue;
import static org.springframework.restdocs.hypermedia.HypermediaDocumentation.linkWithRel;
import static org.springframework.restdocs.hypermedia.HypermediaDocumentation.links;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.documentationConfiguration;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.*;
import static org.springframework.restdocs.payload.PayloadDocumentation.*;
import static org.springframework.test.web.servlet.result.MockMvcResultHandlers.print;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

@RunWith(SpringRunner.class)
@SpringBootTest
public class ApiDocumentation {

	@Rule
	public final JUnitRestDocumentation restDocumentation = new JUnitRestDocumentation();

	@Autowired
	private VilleRepository villeRepository;

	@Autowired
	private InformationRepository informationRepository;

	@Autowired
	private ObjectMapper objectMapper;

	@Autowired
	private WebApplicationContext context;

	private MockMvc mockMvc;

	@Before
	public void setUp() {
		this.mockMvc = MockMvcBuilders.webAppContextSetup(this.context)
				.apply(documentationConfiguration(this.restDocumentation)).build();
	}

	@Test
	public void errorExample() throws Exception {
		this.mockMvc
				.perform(get("/error")
						.requestAttr(RequestDispatcher.ERROR_STATUS_CODE, 400)
						.requestAttr(RequestDispatcher.ERROR_REQUEST_URI,
								"/villes")
						.requestAttr(RequestDispatcher.ERROR_MESSAGE,
								"L'information 'http://localhost:8080/informations/123' n'existe pas"))
				.andDo(print()).andExpect(status().isBadRequest())
				.andExpect(jsonPath("error", is("Bad Request")))
				.andExpect(jsonPath("timestamp", is(notNullValue())))
				.andExpect(jsonPath("status", is(400)))
				.andExpect(jsonPath("path", is(notNullValue())))
				.andDo(document("error-example",
						responseFields(
								fieldWithPath("error").description("L'erreur HTTP qui se produit, i.e. `Bad Request`"),
								fieldWithPath("message").description("Description de la cause de l'erreur"),
								fieldWithPath("path").description("Le path avec lequel la requête a été faite"),
								fieldWithPath("status").description("Le code status HTTP, i.e. `400`"),
								fieldWithPath("timestamp").description("Le temps, en milisecondes qu'a mis l'erreur à se produire"))));
	}

	@Test
	public void indexExample() throws Exception {
		this.mockMvc.perform(get("/"))
			.andExpect(status().isOk())
			.andDo(document("index-example",
					links(
							linkWithRel("villes").description("Les villes <<resources-villes,Villes resource>>"),
							linkWithRel("informations").description("les informations <<resources-informations,Informations resource>>"),
							linkWithRel("profile").description("Le profile ALPS pour le service")),
					responseFields(
							subsectionWithPath("_links").description("<<resources-index-links,Links>> vers les autres ressources"))));

	}

	@Test
	public void villesListExample() throws Exception {
		this.villeRepository.deleteAll();

		createVille("Paris",
				"Capitale de la France");
		createVille("Rouen",
				"C'est est une commune du Nord-Ouest de la France");
		createVille("Marseille", "chef-lieu du département des Bouches-du-Rhône et de la région Sud Provence-Alpes-Côte d'Azur.");

		this.mockMvc.perform(get("/villes"))
			.andExpect(status().isOk())
			.andDo(document("villes-list-example",
					links(
							linkWithRel("self").description("Lien canonique pour cette ressource"),
							linkWithRel("profile").description("Le profile ALPS pour cette ressource")),
					responseFields(
							subsectionWithPath("_embedded.villes").description("Une liste de <<resources-ville, Ville resources>>"),
							subsectionWithPath("_links").description("<<resources-informations-list-links, Links>> vers d'autres ressources"))));
	}

	@Test
	public void villesCreateExample() throws Exception {
		Map<String, String> information = new HashMap<String, String>();
		information.put("nom", "Grand");

		String informationLocation = this.mockMvc
				.perform(
						post("/informations").contentType(MediaTypes.HAL_JSON).content(
								this.objectMapper.writeValueAsString(information)))
				.andExpect(status().isCreated()).andReturn().getResponse()
				.getHeader("Location");

		Map<String, Object> ville = new HashMap<String, Object>();
		ville.put("nom", "Paris");
		ville.put("body", "Capitale de la France");
		ville.put("informations", Arrays.asList(informationLocation));

		this.mockMvc.perform(
				post("/villes").contentType(MediaTypes.HAL_JSON).content(
						this.objectMapper.writeValueAsString(ville))).andExpect(
				status().isCreated())
				.andDo(document("villes-create-example",
						requestFields(
									fieldWithPath("nom").description("Le nom de la ville"),
									fieldWithPath("body").description("La description de la ville"),
									fieldWithPath("informations").description("Une liste d'information"))));
	}

	@Test
	public void villeGetExample() throws Exception {
		Map<String, String> information = new HashMap<String, String>();
		information.put("nom", "Paris");

		String informationLocation = this.mockMvc
				.perform(
						post("/informations").contentType(MediaTypes.HAL_JSON).content(
								this.objectMapper.writeValueAsString(information)))
				.andExpect(status().isCreated()).andReturn().getResponse()
				.getHeader("Location");

		Map<String, Object> ville = new HashMap<String, Object>();
		ville.put("nom", "Paris");
		ville.put("body", "Capitale de la France");
		ville.put("informations", Arrays.asList(informationLocation));

		String villeLocation = this.mockMvc
				.perform(
						post("/villes").contentType(MediaTypes.HAL_JSON).content(
								this.objectMapper.writeValueAsString(ville)))
				.andExpect(status().isCreated()).andReturn().getResponse()
				.getHeader("Location");

		this.mockMvc.perform(get(villeLocation))
			.andExpect(status().isOk())
			.andExpect(jsonPath("nom", is(ville.get("nom"))))
			.andExpect(jsonPath("body", is(ville.get("body"))))
			.andExpect(jsonPath("_links.self.href", is(villeLocation)))
			.andExpect(jsonPath("_links.informations", is(notNullValue())))
			.andDo(print())
			.andDo(document("ville-get-example",
					links(
							linkWithRel("self").description("Len canonique vers cette <<resources-ville,ville>>"),
							linkWithRel("ville").description("Cette <<resources-ville,ville>>"),
							linkWithRel("informations").description("Les informations de cette ville")),
					responseFields(
							fieldWithPath("nom").description("Le nom de la ville"),
							fieldWithPath("body").description("La description de la ville"),
							subsectionWithPath("_links").description("<<resources-ville-links,Links>> vers d'autres ressources"))));
	}

	@Test
	public void informationsListExample() throws Exception {
		this.villeRepository.deleteAll();
		this.informationRepository.deleteAll();

		createInformation("Paris");
		createInformation("Rouen");
		createInformation("Marseille");

		this.mockMvc.perform(get("/informations"))
			.andExpect(status().isOk())
			.andDo(document("informations-list-example",
					links(
							linkWithRel("self").description("Lien canonique pour cette ressource"),
							linkWithRel("profile").description("Le profile ALPS de cette ressource")),
					responseFields(
							subsectionWithPath("_embedded.informations").description("Une liste de <<resources-information,Information resources>>"),
							subsectionWithPath("_links").description("<<resources-informations-list-links, Links>> vers d'autres ressources"))));
	}

	@Test
	public void informationsCreateExample() throws Exception {
		Map<String, String> information = new HashMap<String, String>();
		information.put("nom", "Paris");

		this.mockMvc.perform(
				post("/informations").contentType(MediaTypes.HAL_JSON).content(
						this.objectMapper.writeValueAsString(information)))
				.andExpect(status().isCreated())
				.andDo(document("informations-create-example",
						requestFields(
								fieldWithPath("nom").description("Le nom de l'information"))));
	}

	@Test
	public void villeUpdateExample() throws Exception {
		Map<String, Object> ville = new HashMap<String, Object>();
		ville.put("nom", "Paris");
		ville.put("body", "Capitale de la France");

		String villeLocation = this.mockMvc
				.perform(
						post("/villes").contentType(MediaTypes.HAL_JSON).content(
								this.objectMapper.writeValueAsString(ville)))
				.andExpect(status().isCreated()).andReturn().getResponse()
				.getHeader("Location");

		this.mockMvc.perform(get(villeLocation)).andExpect(status().isOk())
				.andExpect(jsonPath("nom", is(ville.get("nom"))))
				.andExpect(jsonPath("body", is(ville.get("body"))))
				.andExpect(jsonPath("_links.self.href", is(villeLocation)))
				.andExpect(jsonPath("_links.informations", is(notNullValue())));

		Map<String, String> information = new HashMap<String, String>();
		information.put("nom", "Paris");

		String informationLocation = this.mockMvc
				.perform(
						post("/informations").contentType(MediaTypes.HAL_JSON).content(
								this.objectMapper.writeValueAsString(information)))
				.andExpect(status().isCreated()).andReturn().getResponse()
				.getHeader("Location");

		Map<String, Object> villeUpdate = new HashMap<String, Object>();
		villeUpdate.put("informations", Arrays.asList(informationLocation));

		this.mockMvc.perform(
				patch(villeLocation).contentType(MediaTypes.HAL_JSON).content(
						this.objectMapper.writeValueAsString(villeUpdate)))
				.andExpect(status().isNoContent())
				.andDo(document("ville-update-example",
						requestFields(
								fieldWithPath("nom").description("Le nom de la ville").type(JsonFieldType.STRING).optional(),
								fieldWithPath("body").description("La description de la ville").type(JsonFieldType.STRING).optional(),
								fieldWithPath("informations").description("Un liste d'informations").optional())));
	}

	@Test
	public void informationGetExample() throws Exception {
		Map<String, String> information = new HashMap<String, String>();
		information.put("nom", "Paris");

		String informationLocation = this.mockMvc
				.perform(
						post("/informations").contentType(MediaTypes.HAL_JSON).content(
								this.objectMapper.writeValueAsString(information)))
				.andExpect(status().isCreated()).andReturn().getResponse()
				.getHeader("Location");

		this.mockMvc.perform(get(informationLocation))
			.andExpect(status().isOk())
			.andExpect(jsonPath("nom", is(information.get("nom"))))
			.andDo(document("information-get-example",
					links(
							linkWithRel("self").description("Le nom canonique de cette  <<resources-information,information>>"),
							linkWithRel("information").description("Cette <<resources-information,information>>"),
							linkWithRel("villes").description("La <<resources-informationged-villes,villes>> qui a cette information")),
					responseFields(
							fieldWithPath("nom").description("Le nom de l'information"),
							subsectionWithPath("_links").description("<<resources-information-links,Links>> vers d'autres ressources"))));
	}

	@Test
	public void informationUpdateExample() throws Exception {
		Map<String, String> information = new HashMap<String, String>();
		information.put("nom", "Paris");

		String informationLocation = this.mockMvc
				.perform(
						post("/informations").contentType(MediaTypes.HAL_JSON).content(
								this.objectMapper.writeValueAsString(information)))
				.andExpect(status().isCreated()).andReturn().getResponse()
				.getHeader("Location");

		Map<String, Object> informationUpdate = new HashMap<String, Object>();
		informationUpdate.put("nom", "Rouen");

		this.mockMvc.perform(
				patch(informationLocation).contentType(MediaTypes.HAL_JSON).content(
						this.objectMapper.writeValueAsString(informationUpdate)))
				.andExpect(status().isNoContent())
				.andDo(document("information-update-example",
						requestFields(
								fieldWithPath("nom").description("Le nom de l'information"))));
	}

	private void createVille(String nom, String body) {
		Ville ville = new Ville();
		ville.setNom(nom);
		ville.setBody(body);

		this.villeRepository.save(ville);
	}

	private void createInformation(String nom) {
		Information information = new Information();
		information.setNom(nom);
		this.informationRepository.save(information);
	}
}
