package fr.eni.ep3jasp.doc;

import static org.hamcrest.Matchers.hasSize;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.Matchers.notNullValue;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.documentationConfiguration;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.get;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.patch;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.post;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.header;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.io.UnsupportedEncodingException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.hateoas.MediaTypes;
import org.springframework.restdocs.JUnitRestDocumentation;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;
import org.springframework.web.context.WebApplicationContext;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.jayway.jsonpath.JsonPath;

@RunWith(SpringRunner.class)
@SpringBootTest
public class GettingStartedDocumentation {

	@Rule
	public final JUnitRestDocumentation restDocumentation = new JUnitRestDocumentation();

	@Autowired
	private ObjectMapper objectMapper;

	@Autowired
	private WebApplicationContext context;

	private MockMvc mockMvc;

	@Before
	public void setUp() {
		this.mockMvc = MockMvcBuilders.webAppContextSetup(this.context)
				.apply(documentationConfiguration(this.restDocumentation))
				.alwaysDo(document("{method-name}/{step}/"))
				.build();
	}

	@Test
	public void index() throws Exception {
		this.mockMvc.perform(get("/").accept(MediaTypes.HAL_JSON))
			.andExpect(status().isOk())
			.andExpect(jsonPath("_links.villes", is(notNullValue())))
			.andExpect(jsonPath("_links.informations", is(notNullValue())));
	}

	@Test
	public void creatingAVille() throws JsonProcessingException, Exception {
		String villeLocation = createVille();
		MvcResult ville = getVille(villeLocation);

		String informationLocation = createInformation();
		getInformation(informationLocation);

		String informationgedVilleLocation = createInformationgedVille(informationLocation);
		MvcResult informationgedVille = getVille(informationgedVilleLocation);
		getInformations(getLink(informationgedVille, "informations"));

		informationExistingVille(villeLocation, informationLocation);
		getInformations(getLink(ville, "informations"));
	}

	String createVille() throws Exception {
		Map<String, String> ville = new HashMap<String, String>();
		ville.put("nom", "Ville creation with cURL");
		ville.put("body", "An example of how to create a ville using cURL");

		String villeLocation = this.mockMvc
				.perform(
						post("/villes").contentType(MediaTypes.HAL_JSON).content(
								objectMapper.writeValueAsString(ville)))
				.andExpect(status().isCreated())
				.andExpect(header().string("Location", notNullValue()))
				.andReturn().getResponse().getHeader("Location");
		return villeLocation;
	}

	MvcResult getVille(String villeLocation) throws Exception {
		return this.mockMvc.perform(get(villeLocation))
				.andExpect(status().isOk())
				.andExpect(jsonPath("nom", is(notNullValue())))
				.andExpect(jsonPath("body", is(notNullValue())))
				.andExpect(jsonPath("_links.informations", is(notNullValue())))
				.andReturn();
	}

	String createInformation() throws Exception, JsonProcessingException {
		Map<String, String> information = new HashMap<String, String>();
		information.put("nom", "getting-started");

		String informationLocation = this.mockMvc
				.perform(
						post("/informations").contentType(MediaTypes.HAL_JSON).content(
								objectMapper.writeValueAsString(information)))
				.andExpect(status().isCreated())
				.andExpect(header().string("Location", notNullValue()))
				.andReturn().getResponse().getHeader("Location");
		return informationLocation;
	}

	void getInformation(String informationLocation) throws Exception {
		this.mockMvc.perform(get(informationLocation)).andExpect(status().isOk())
			.andExpect(jsonPath("nom", is(notNullValue())))
			.andExpect(jsonPath("_links.villes", is(notNullValue())));
	}

	String createInformationgedVille(String information) throws Exception {
		Map<String, Object> ville = new HashMap<String, Object>();
		ville.put("nom", "Informationged ville creation with cURL");
		ville.put("body", "An example of how to create a informationged ville using cURL");
		ville.put("informations", Arrays.asList(information));

		String villeLocation = this.mockMvc
				.perform(
						post("/villes").contentType(MediaTypes.HAL_JSON).content(
								objectMapper.writeValueAsString(ville)))
				.andExpect(status().isCreated())
				.andExpect(header().string("Location", notNullValue()))
				.andReturn().getResponse().getHeader("Location");
		return villeLocation;
	}

	void getInformations(String villeInformationsLocation) throws Exception {
		this.mockMvc.perform(get(villeInformationsLocation))
			.andExpect(status().isOk())
			.andExpect(jsonPath("_embedded.informations", hasSize(1)));
	}

	void informationExistingVille(String villeLocation, String informationLocation) throws Exception {
		Map<String, Object> update = new HashMap<String, Object>();
		update.put("informations", Arrays.asList(informationLocation));

		this.mockMvc.perform(
				patch(villeLocation).contentType(MediaTypes.HAL_JSON).content(
						objectMapper.writeValueAsString(update)))
				.andExpect(status().isNoContent());
	}

	MvcResult getInformationgedExistingVille(String villeLocation) throws Exception {
		return this.mockMvc.perform(get(villeLocation))
			.andExpect(status().isOk())
			.andReturn();
	}

	void getInformationsForExistingVille(String villeInformationsLocation) throws Exception {
		this.mockMvc.perform(get(villeInformationsLocation))
			.andExpect(status().isOk())
			.andExpect(jsonPath("_embedded.informations", hasSize(1)));
	}

	private String getLink(MvcResult result, String rel)
			throws UnsupportedEncodingException {
		return JsonPath.parse(result.getResponse().getContentAsString()).read(
				"_links." + rel + ".href");
	}
}
