package fr.eni.ep3jasp.mongodb.ex2;

import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.MvcResult;

import static org.hamcrest.Matchers.containsString;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultHandlers.print;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

// tag::test1[]
@RunWith(SpringRunner.class)
@SpringBootTest
@AutoConfigureMockMvc
public class Ex2MongoDBTests {
  @Autowired
  private MockMvc mockMvc;
  @Autowired
  private CapitaleRepository capitaleRepository;
  @Before
  public void deleteAllBeforeTests() throws Exception {
    capitaleRepository.deleteAll();
  }

// end::test1[]

  // tag::creation[]
  @Test
  public void shouldCreateCapitale() throws Exception {
    mockMvc.perform(post("/capitales").content(
          "{\"pays\": \"Haïti\", \"capitale\":\"Port-au-Prince\", \"continent\":\"Amérique\"}")).andExpect(
          status().isCreated()).andExpect(
          header().string("Location", containsString("capitales/")));
  }
  // end::creation[]

  // tag::lecture[]
  @Test
  public void shouldRetrieveCapitale() throws Exception {
    MvcResult mvcResult = mockMvc.perform(post("/capitales").content(
          "{\"pays\": \"Haïti\", \"capitale\":\"Port-au-Prince\", \"continent\":\"Amérique\"}")).andExpect(
          status().isCreated()).andReturn();
    String location = mvcResult.getResponse().getHeader("Location");
    mockMvc.perform(get(location)).andExpect(status().isOk()).andExpect(
          jsonPath("$.pays").value("Haïti")).andExpect(
          jsonPath("$.capitale").value("Port-au-Prince")).andExpect(
          jsonPath("$.continent").value("Amérique"));
  }
  // end::lecture[]

  // tag::recherche[]
  @Test
  public void shouldQueryCapitale() throws Exception {
    MvcResult mvcResult = mockMvc.perform(post("/capitales").content(
          "{\"pays\": \"Haïti\", \"capitale\":\"Port-au-Prince\", \"continent\":\"Amérique\"}")).andExpect(
          status().isCreated()).andReturn();
    mockMvc.perform(
          get("/capitales/search/findByContinent?continent={continent}", "Amérique")).andExpect(
          status().isOk()).andDo(print()).andExpect(
          jsonPath("$._embedded.capitale[0].pays").value(
                "Haïti"));
  }
  // end::recherche[]

  // tag::miseajour[]
  @Test
  public void shouldUpdateCapitale() throws Exception {
    MvcResult mvcResult = mockMvc.perform(post("/capitales").content(
          "{\"pays\": \"Haïti\", \"capitale\":\"Port-au-Prince\", \"continent\":\"Amérique\"}")).andExpect(
          status().isCreated()).andReturn();
    String location = mvcResult.getResponse().getHeader("Location");
    mockMvc.perform(put(location).content(
          "{\"pays\": \"Haïti\", \"capitale\":\"Port-au-Prince\", \"continent\":\"Amérique\"}")).andExpect(
          status().isNoContent());
    mockMvc.perform(get(location)).andExpect(status().isOk()).andExpect(
          jsonPath("$.pays").value("Haïti")).andExpect(
          jsonPath("$.capitale").value("Port-au-Prince")).andExpect(
          jsonPath("$.continent").value("Amérique"));
  }
  // end::miseajour[]

  // tag::miseajourpartielle[]
  @Test
  public void shouldPartiallyUpdateCapitale() throws Exception {
    MvcResult mvcResult = mockMvc.perform(post("/capitales").content(
          "{\"pays\": \"Haïti\", \"capitale\":\"Port-au-Prince\", \"continent\":\"Amérique\"}")).andExpect(
          status().isCreated()).andReturn();
    String location = mvcResult.getResponse().getHeader("Location");
    mockMvc.perform(
          patch(location).content("{\"pays\": \"HAITI\"}")).andExpect(
          status().isNoContent());
    mockMvc.perform(get(location)).andExpect(status().isOk()).andExpect(
          jsonPath("$.pays").value("HAITI")).andExpect(
          jsonPath("$.capitale").value("Port-au-Prince")).andExpect(
          jsonPath("$.continent").value("Amérique"));
  }
  // end::miseajourpartielle[]

  // tag::suppr[]
  @Test
  public void shouldDeleteCapitale() throws Exception {
    MvcResult mvcResult = mockMvc.perform(post("/capitales").content(
          "{\"pays\": \"Haïti\", \"capitale\":\"Port-au-Prince\", \"continent\":\"Amérique\"}")).andExpect(
          status().isCreated()).andReturn();
    String location = mvcResult.getResponse().getHeader("Location");
    mockMvc.perform(delete(location)).andExpect(status().isNoContent());
    mockMvc.perform(get(location)).andExpect(status().isNotFound());
  }
  // end::suppr[]
}